/* Name: cdc162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"
#include "msc162.h"

#define USE_SERFLASH
#ifdef USE_SERFLASH
#include "serflash.h"
#define SERFLASH_SECTOR_SHIFT	 9 // 512 B
//#define SERFLASH_SECTOR_SHIFT	12 // 4KB
#define SERFLASH_SECTOR_SIZE	(1<<SERFLASH_SECTOR_SHIFT)
#endif

#ifdef EP_MS_OUT
/* ---- */
static void usbmsc_poll() {
	UENUM = EP_MS_OUT;
	if (bit_is_set(UEINTX,RXOUTI)) {
		bit_clear(UEINTX, RXOUTI);
	}
	if (bit_is_set(UEINTX, FIFOCON) && bit_is_clear(UEINTX,RWAL)) {
		bit_clear(UEINTX, FIFOCON);
	}
	UENUM = EP_MS_IN;
	if (bit_is_set(UEINTX,TXINI)) {
		bit_clear(UEINTX,TXINI);
		bit_clear(UEINTX,NAKINI);
	}
#if 0
	if (bit_is_set(UDINT, SOFI)) {
		if (UEBCLX) bit_set(GPIOR2, USB162_RX_SEND_REQ2);
	}
	if (bit_is_set(UEINTX, FIFOCON)
		&& (bit_is_set(GPIOR2, USB162_RX_SEND_EMPTY2)
		    || bit_is_clear(UEINTX,RWAL)
		    || bit_is_set(GPIOR2,USB162_RX_SEND_REQ2))) {
		bit_clear(GPIOR2, USB162_RX_SEND_REQ2);
		bit_clear(GPIOR2, USB162_RX_SEND_EMPTY2);
		if (UEBCLX) bit_set(GPIOR2, USB162_RX_SEND_EMPTY2);
		bit_clear(UEINTX, FIFOCON);
	}
#else
	if (bit_is_set(UEINTX, FIFOCON)
		&& (bit_is_clear(UEINTX,RWAL) || bit_is_set(GPIOR2,USB162_RX_SEND_REQ2))) {
		bit_clear(GPIOR2, USB162_RX_SEND_REQ2);
		bit_clear(UEINTX, FIFOCON);
	}
#endif

	usbPoll();
}

static int8_t usbmsc_can_getc(int8_t size) {
	UENUM = EP_MS_OUT;
	return (UEBCLX >= size);
}

static int8_t usbmsc_can_putc(int8_t size) {
	UENUM = EP_MS_IN;
	return (UEBCLX + size) <= EP_MS_IN_SIZE;
}

static unsigned char usbmsc_getc(void) {
	unsigned char ret;
	UENUM = EP_MS_OUT;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbmsc_poll();
		UENUM = EP_MS_OUT;
	}
	ret = UEDATX;
	return ret;
}

static void usbmsc_putc(unsigned char data) {
	UENUM = EP_MS_IN;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbmsc_poll();
		UENUM = EP_MS_IN;
	}
	UEDATX = data;
}

static inline void usbmsc_oflush() {
	bit_set(GPIOR2, USB162_RX_SEND_REQ2);
}

/* ---- */

#define MSC_LOG_SIZE	0xc0

uint8_t msc_log[MSC_LOG_SIZE];
uint8_t msc_log_idx;

void msc_debug(uint8_t ch) {
	if (msc_log_idx < MSC_LOG_SIZE) {
		msc_log[msc_log_idx++] = ch;

	}
#if 0
	else if (msc_log_idx == MSC_LOG_SIZE) {
		    eeprom_write_block(msc_log, (void *)0, MSC_LOG_SIZE);
	}
#endif
}

#define SC_VENDOR_ID	'A','T','M','E','L',' ',' ',' '
#define SC_PRODUCT_ID	'A','T','9','0','U','S','B','1','2','8',' ','M',' ','S',' ',' '
#define SC_REVISION_ID	'0','.','0','0'

uint8_t msc_inquiry_data[] PROGMEM = {
	0x00, // DeviceType: Direct-access device , 0x05: CDROM
	0x80, // Medium is removable
	0x03, // Version : SPC (=SBC) (ANSI SCSI level 3 ?)
	0x02, // Response data format
	31,   // Additional Length
	0,0,0, // Reserved
	SC_VENDOR_ID,
	SC_PRODUCT_ID,
	SC_REVISION_ID,
};

uint8_t msc_maxlun;
uint32_t msc_capacity = 0;
msc_cbw_t msc_cbw;

#define MSC_RESULT_OK		0
#define MSC_RESULT_CMD_ERR	1
#define MSC_RESULT_FRAME_ERR	2

uint8_t msc_sense_key = SC_SENSE_KEY_NO_SENSE;
uint8_t msc_sense_asc = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;

volatile uint32u_t msc_trans_offset;
void (* msc_trans_proc)();

#define MSC_STATE_COMMAND	0
#define MSC_STATE_TRANS		1
#define MSC_STATE_STATUS	2

uint8_t msc_state = MSC_STATE_COMMAND;

static void usbmsc_putc2(uint8_t ch) {
	usbmsc_putc(0); 
	usbmsc_putc(ch); 
}

static void usbmsc_putc4(uint8_t ch) {
	usbmsc_putc2(0); 
	usbmsc_putc2(ch); 
}

static void usbmsc_putc2_same(uint8_t ch) {
	usbmsc_putc(ch); 
	usbmsc_putc(ch); 
}


void callback_msc_reset(uint8_t hard) {
	/* endpoint FIFO Reset */
	bit_set(UERST , EP_MS_IN);
	bit_clear(UERST , EP_MS_IN);
	bit_set(UERST , EP_MS_OUT);
	bit_clear(UERST , EP_MS_OUT);
	if (hard) {
		msc_debug(0xfe);
	} else {
		msc_debug(0xfd);
	}
	msc_state = MSC_STATE_COMMAND;
}

uint8_t callback_msc_maxlun() {
	msc_debug(0xfc);
	return msc_maxlun;
}

static void msc_trans_skip() {
	if (msc_cbw.cbw.trans_len.u32 == 0L) return;
	if (msc_cbw.cbw.flags & 0x80) { /* Bulk IN */
	    if (usbmsc_can_putc(1)) {
		while (usbmsc_can_putc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
			usbmsc_putc(0);
			msc_cbw.cbw.trans_len.u32--;
		}
	    }
	} else { /* Bulk OUT */
		while (usbmsc_can_getc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
			usbmsc_getc(); 
			msc_cbw.cbw.trans_len.u32--;
		}
	}
}

static void msc_trans_inquiry() {
	while (usbmsc_can_putc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
		usbmsc_putc(__LPM(msc_inquiry_data+msc_trans_offset.ub[0]));
		msc_trans_offset.ub[0]++;
		msc_cbw.cbw.trans_len.u32 --;
		if (msc_trans_offset.ub[0] >= 36) {
	    		msc_trans_proc = msc_trans_skip;
			break;
		}
	}
}

static void msc_trans_request_sense() {
	if (usbmsc_can_putc(18)) {
		usbmsc_putc(0x70); // RESPONSE_CODE_SENSE
		usbmsc_putc(0x00);
		usbmsc_putc(msc_sense_key);
		usbmsc_putc4(0x00);
		usbmsc_putc(10); // ADDITIONAL_SENSE_LENGTH
		usbmsc_putc4(0x00);
		usbmsc_putc(msc_sense_asc);
		usbmsc_putc(0x00); // msc_sense_ascq
		usbmsc_putc4(0x00);
		msc_cbw.cbw.trans_len.u32 -= 18;
		msc_sense_key = SC_SENSE_KEY_NO_SENSE;
		msc_sense_asc = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;
		msc_trans_proc = msc_trans_skip;
	}
}

#define MSC_SENSE_MODE10		(1<<0)
#define MSC_SENSE_PAGE_CODE_ERR_6	(1<<1)
#define MSC_SENSE_PAGE_CODE_ERR_10	(1<<2)
#define MSC_SENSE_PAGE_CODE_EXPT	(1<<3)
#define MSC_SENSE_PAGE_CODE_CACHING	(1<<4)

uint8_t msc_sense_mode;
uint8_t msc_pagecode_len;
uint8_t msc_write_protected;

static void msc_trans_mode_sense() {
	if (usbmsc_can_putc(32)) {
	    if (msc_sense_mode & MSC_SENSE_MODE10) {
		usbmsc_putc2(msc_pagecode_len); // Data length
		usbmsc_putc(0x00); // DeviceType: DEFAULT_MEDIUM_TYPE
		usbmsc_putc(msc_write_protected ? 0x80:0x00);
		usbmsc_putc4(0x00); // BLOCK_DESCRIPTOR_LENGTH
		msc_cbw.cbw.trans_len.u32 -= 8;
	    } else {
		usbmsc_putc(msc_pagecode_len); // Data length
		usbmsc_putc(0x00); // DeviceType: DEFAULT_MEDIUM_TYPE
		usbmsc_putc(msc_write_protected ? 0x80:0x00);
		usbmsc_putc(0x00); // BLOCK_DESCRIPTOR_LENGTH
		msc_cbw.cbw.trans_len.u32 -= 4;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_ERR_6) {
		usbmsc_putc(SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY);
		usbmsc_putc(6);    // PAGE_LENGTH
		usbmsc_putc(0x80);
		usbmsc_putc(0x03);  // READ_RETRY_COUNT
		usbmsc_putc4(0x00); // CORRECTION_SPAN,HEAD_OFFSET_COUNT
				    // ,DATA_STROBE_OFFSET,Reserved
		msc_cbw.cbw.trans_len.u32 -= 8;
	    }
	    else if (msc_sense_mode & MSC_SENSE_PAGE_CODE_ERR_10) {
		usbmsc_putc(SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x80);
		usbmsc_putc(0x03);  // READ_RETRY_COUNT
		usbmsc_putc4(0x00); // CORRECTION_SPAN,HEAD_OFFSET_COUNT
				    // ,DATA_STROBE_OFFSET,Reserved

		usbmsc_putc(0x80);  // WRITE_RETRY_COUNT
		usbmsc_putc(0x00);
		usbmsc_putc2(0x00); // RECOVERY_LIMIT
		msc_cbw.cbw.trans_len.u32 -= 12;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_EXPT) {
		// Page Code: Informational exceptions control page
		// See chapter 8.3.8 on SPC-2 specification
		usbmsc_putc(SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x00);
		usbmsc_putc(0x05);  // MRIE
		usbmsc_putc4(0x00); // Interval Timer 
		usbmsc_putc4(0x01); // Report Count
		msc_cbw.cbw.trans_len.u32 -= 12;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_CACHING) {
		usbmsc_putc(SC_PAGE_CODE_CACHING_MODE_PAGE);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x04);  // Write Cache Enable , Read Cache Disabled
		usbmsc_putc(0x00);
		usbmsc_putc2_same(0xff); // don't disable prefetch
		usbmsc_putc2(0x00);
		usbmsc_putc2_same(0xff); //  Maximum prefetch
		usbmsc_putc2_same(0xff); // Maximum prefetch ceiling
		msc_cbw.cbw.trans_len.u32 -= 12;
	    }
	    msc_trans_proc = msc_trans_skip;
	}
}

static void msc_trans_read_capacity() {
	if (usbmsc_can_putc(8)) {
		usbmsc_putc((uint8_t)(msc_capacity>>24));
		usbmsc_putc((uint8_t)(msc_capacity>>16));
		usbmsc_putc((uint8_t)(msc_capacity>>8));
		usbmsc_putc((uint8_t)(msc_capacity));
		usbmsc_putc2(0x00); // SECTOR SIZE
		usbmsc_putc(SERFLASH_SECTOR_SIZE >> 8);
		usbmsc_putc(SERFLASH_SECTOR_SIZE & 0xff);
		msc_cbw.cbw.trans_len.u32 -= 8;
		msc_trans_proc = msc_trans_skip;
	}
}

static void msc_trans_read() {
	while (usbmsc_can_putc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
		usbmsc_putc(serflash_proc(0));
		msc_cbw.cbw.trans_len.u32--;
	}
	if (msc_cbw.cbw.trans_len.u32 == 0L) {
		serflash_close();
		return;
	}
}

#define MSC_WRITE_STAT_INIT	0
#define MSC_WRITE_STAT_PGM	1
#define MSC_WRITE_STAT_ERASE	2

static uint8_t msc_write_stat;

static void msc_trans_write() {
	while (usbmsc_can_getc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
		if (msc_trans_offset.ub[0] == 0) {
			serflash_close();
			if (serflash_status() & SERFLASH_STAT_WIP) {
				return;
			}
			msc_debug(0xb0);
			if ((msc_trans_offset.ub[1] & 0x0f) == 0) {
				if (msc_write_stat != MSC_WRITE_STAT_ERASE) {
					msc_debug(0xb1);
					msc_write_stat = MSC_WRITE_STAT_ERASE;
					serflash_erase_4k(msc_trans_offset.u32);
					return;
				}
				msc_debug(0xb2);
			}
			msc_write_stat = MSC_WRITE_STAT_PGM;
			serflash_write_enable();
			serflash_open(SERFLASH_PAGE_PGM);
			serflash_seek(msc_trans_offset.u32, SEEK_SET);
		}
		serflash_proc(usbmsc_getc()); 
		msc_cbw.cbw.trans_len.u32--;
		msc_trans_offset.u32++;
	}
	if (msc_cbw.cbw.trans_len.u32 == 0L) {
		if (msc_write_stat != MSC_WRITE_STAT_INIT) {
			msc_write_stat = MSC_WRITE_STAT_INIT;
			serflash_close();
			msc_debug(0xb3);
		}
		return;
	}
}
#endif

void msc_init() {
#ifdef EP_MS_OUT
  #if 0
	eeprom_write_block(msc_log, (void *)0, MSC_LOG_SIZE);
  #endif
#ifdef USE_SERFLASH
	serflash_init();
	msc_capacity = (serflash_size >> SERFLASH_SECTOR_SHIFT) - 1;
#endif
#endif
}

void msc_print_status() {
#if defined(EP_TX) && defined(EP_MS_OUT)
	uint8_t i,j,d;
	uint16_t offset = 0;
	usbcdc_puts("msc ");
	usbcdc_putc('0' + msc_state);
	usbcdc_puts("\r\n");
	for(i=0; i<12; i++) {
		d = (offset >> 8) & 0xff;
		usbcdc_put_hex(d);
		d = offset & 0xff;
		usbcdc_put_hex(d);
		usbcdc_putc(':');
		for (j=0; j<16; j++) {
			usbcdc_putc(' ');
			//d = eeprom_read_byte((void *)offset);
			d = msc_log[offset];
			usbcdc_put_hex(d);
			offset++;
		}
		usbcdc_puts("\r\n");
	}
	msc_log_idx = 0;
#endif
}

void msc_poll() {
#ifdef EP_MS_OUT
	uint8_t i;
	uint8_t sc_cmd;
	uint8_t sc_len = 14;
	uint8_t sc_lun;
	uint8_t wk;
	switch (msc_state) {
	case MSC_STATE_COMMAND:
		if (usbmsc_can_getc(31)) {
			for (i=0; i<15; i++) {
				msc_cbw.ub[i] = usbmsc_getc();
			}
			msc_trans_proc = msc_trans_skip;
			msc_trans_offset.u32 = 0L;
			sc_cmd = usbmsc_getc();
			sc_lun = usbmsc_getc() >> 5;
			msc_debug(0xc0 + sc_lun);
			msc_debug(sc_cmd);
			msc_debug(msc_cbw.cbw.trans_len.ub[0]);

			if (sc_cmd != SC_REQUEST_SENSE) {
				msc_sense_key = SC_SENSE_KEY_NO_SENSE;
				msc_sense_asc = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;
			}
			if (sc_lun > msc_maxlun) {
				msc_sense_key = SC_SENSE_KEY_NOT_READY;
				msc_sense_asc = SC_ASC_LOGICAL_UNIT_NOT_READY;
			} else switch (sc_cmd) {
			case SC_INQUIRY:
				msc_trans_proc = msc_trans_inquiry;
				break;
			case SC_MODE_SENSE_6:
				wk = usbmsc_getc(); // pagecode
				sc_len -= 1;
				msc_debug(wk);
				msc_trans_proc = msc_trans_mode_sense;
				msc_sense_mode = 0;
				msc_pagecode_len = 3;
				if (msc_cbw.cbw.trans_len.ub[0] < 12) {
					break;
				}
				switch ( wk & SC_MSK_PAGE_CODE ) {
				case SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY:
					if (msc_cbw.cbw.trans_len.ub[0] < 16) {
						msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_6;
						msc_pagecode_len += 8;
						break;
					}
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_10;
					msc_pagecode_len += 12;
					break;
				case SC_PAGE_CODE_CACHING_MODE_PAGE:
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_CACHING;
					msc_pagecode_len += 12;
					break;
				case SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS: 
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT;
					msc_pagecode_len += 12;
					break;
				case SC_PAGE_CODE_ALL:
					if (msc_cbw.cbw.trans_len.ub[0] < 16) {
						msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_6;
						msc_pagecode_len += 8;
						break;
					}
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_10;
					msc_pagecode_len += 12;
					if (msc_cbw.cbw.trans_len.ub[0] < 28) {
						break;
					}
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT;
					msc_pagecode_len += 12;
					break;
				default:
					//msc_sense_key = SC_SENSE_KEY_ILLEGAL_REQUEST;
					//msc_sense_asc = SC_ASC_INVALID_FIELD_IN_CDB;
					break;
				}
				break;
#if 0
			case SC_MODE_SENSE_10:
				msc_sense_mode = MSC_SENSE_MODE10;
				msc_trans_proc = msc_trans_mode_sense;
				switch ( pagecode & SC_MSK_PAGE_CODE ) {
				case SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS: 
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT;
					msc_pagecode_len = 12 + 3;
					break;
				case SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY:
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR1;
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR2;
					msc_pagecode_len = 12 + 3;
					break;
				case SC_PAGE_CODE_ALL:
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT;
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR1;
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR2;
					msc_pagecode_len = 12 + 12 + 3;
					break;
				default:
					msc_sense_key = SC_SENSE_KEY_ILLEGAL_REQUEST;
					msc_sense_asc = SC_ASC_INVALID_FIELD_IN_CDB;
					msc_trans_proc = msc_trans_skip;
				}
				break;
#endif
			case SC_PREVENT_ALLOW_MEDIUM_REMOVAL:
				break;
			case SC_READ_10:
				msc_trans_offset.ub[3] = usbmsc_getc();
				msc_trans_offset.ub[2] = usbmsc_getc();
				msc_trans_offset.ub[1] = usbmsc_getc();
				msc_trans_offset.ub[0] = usbmsc_getc();
				msc_trans_offset.u32 <<= SERFLASH_SECTOR_SHIFT;
				sc_len -= 4;
				serflash_open(SERFLASH_READ);
				serflash_seek(msc_trans_offset.u32,SEEK_SET);
				msc_trans_proc = msc_trans_read;
				break;
			case SC_WRITE_10:
				msc_trans_offset.ub[3] = usbmsc_getc();
				msc_trans_offset.ub[2] = usbmsc_getc();
				msc_trans_offset.ub[1] = usbmsc_getc();
				msc_trans_offset.ub[0] = usbmsc_getc();
				msc_trans_offset.u32 <<= SERFLASH_SECTOR_SHIFT;
				sc_len -= 4;
				msc_trans_proc = msc_trans_write;
				break;
			case SC_READ_CAPACITY:
				msc_trans_proc = msc_trans_read_capacity;
				break;
			case SC_REQUEST_SENSE:
				msc_trans_proc = msc_trans_request_sense;
				break;
			case SC_TEST_UNIT_READY:
				break;
			case SC_VERIFY_10:
				msc_trans_proc = msc_trans_skip;
				break;
			default:
				msc_sense_key = SC_SENSE_KEY_ILLEGAL_REQUEST;
				msc_sense_asc = SC_ASC_INVALID_COMMAND_OPERATION_CODE;
			}
			for (i=0; i<sc_len; i++) {
				usbmsc_getc();
			}
			if (msc_cbw.cbw.trans_len.u32 == 0L) {
				msc_state = MSC_STATE_STATUS;
			} else {
				msc_state = MSC_STATE_TRANS;
			}
		}
		break;
	case MSC_STATE_TRANS:
		(msc_trans_proc)();
		if (msc_cbw.cbw.trans_len.u32 == 0L) {
		    if (msc_cbw.cbw.flags & 0x80) { // Bulk IN
			usbmsc_oflush();
		    }
		    msc_state = MSC_STATE_STATUS;
		}
		break;
	case MSC_STATE_STATUS:
		if (usbmsc_can_putc(13)) {
			msc_cbw.csw.magic[3] = 'S'; // 'USBC' -> 'USBS'
			msc_cbw.csw.status = 
				(msc_sense_key == SC_SENSE_KEY_NO_SENSE)?0:1;
			for (i=0; i<13; i++) {
				usbmsc_putc(msc_cbw.ub[i]); 
			}
			msc_debug(msc_sense_key);
			msc_state = MSC_STATE_COMMAND;
			usbmsc_oflush();
		}
		break;
	default:
		break;
	}
	usbmsc_poll();
#endif
}
