/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbconfig.h"
#include "cdc162.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )

#include "serflash.h"

void init_spi() {
	bit_clear(TGT_DDR, TGT_MISO);
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_MISO) );
	TGT_DDR |=    ((1<<TGT_SCK) | (1<<TGT_MOSI));
	bit_set(SPI_CS_PORT, SPI_CS);
	bit_set(SPI_CS_DDR, SPI_CS);
}

void term_spi() {
	TGT_DDR &=   ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_MISO));
	SPI_CS_DDR &=   ~(1<<SPI_CS);
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_MISO));
	bit_clear(SPI_CS_DDR, SPI_CS);
	bit_clear(SPI_CS_PORT, SPI_CS);
}

uint8_t spi_write(uint8_t c) {
	uint8_t i,rx;
	i = 0x80;
	rx = 0;
	do {
		if (c & i) {
			bit_set(TGT_PORT, TGT_MOSI);
		} else {
			bit_clear(TGT_PORT, TGT_MOSI);
		}
		bit_set(TGT_PORT, TGT_SCK);
		if (bit_is_set(TGT_PIN, TGT_MISO)) {
			rx |= i;
		}
		bit_clear(TGT_PORT, TGT_SCK);
		i >>= 1;
	} while (i);
	return rx;
}

uint8_t serflash_signature[3];
uint8_t serflash_type;
uint8_t serflash_cmd;
uint32_t serflash_offset;
uint32_t serflash_size;

int8_t serflash_init() {
	spi_enable();
	spi_write(SERFLASH_RDID);
	serflash_signature[0] = spi_read();
	serflash_signature[1] = spi_read();
	serflash_signature[2] = spi_read();
	spi_disable();
	if (serflash_signature[0] == 0x37) {
		if (serflash_signature[1] == 0x30) {
			if (serflash_signature[2] == 0x14) {
				serflash_type = SERFLASH_TYPE_A25L080;
				serflash_size = 1 * 1024 * 1024L;
			}
		}
	} else if (serflash_signature[0] == 0x1f) {
		if (serflash_signature[1] == 0x44) {
				serflash_type = SERFLASH_TYPE_AT25DF041;
				serflash_size = 512 * 1024L;
		} else if (serflash_signature[1] == 0x45) {
				serflash_type = SERFLASH_TYPE_AT25DF081;
				serflash_size = 1 * 1024 * 1024L;
		} else if (serflash_signature[1] == 0x46) {
				serflash_type = SERFLASH_TYPE_AT25DF161;
				serflash_size = 2 * 1024 * 1024L;
		}
	}
	return serflash_type;
}

void serflash_seek(int32_t offs,int8_t whence) {
	if (whence == SEEK_END) serflash_offset = serflash_size + offs;
	else if (whence == SEEK_CUR) serflash_offset += offs;
	else serflash_offset = offs;
	if (serflash_cmd) {
		spi_write(serflash_cmd);
		spi_write((serflash_offset >> 16) & 0xff);
		spi_write((serflash_offset >> 8) & 0xff);
		spi_write(serflash_offset  & 0xff);
	}
}

void serflash_erase_4k(uint32_t offs) {

	serflash_write_enable();

	serflash_open(SERFLASH_SECTOR_ERASE);
	serflash_seek(offs,SEEK_SET);
	serflash_close();
}

